==> release-it-14.1.0-next.0.tgz: package/lib/plugin/npm/npm.js <==
const path = require('path');
const semver = require('semver');
const urlJoin = require('url-join');
const Plugin = require('../Plugin');
const { hasAccess, rejectAfter, parseVersion, e } = require('../../util');
const prompts = require('./prompts');

const docs = 'https://git.io/release-it-npm';

const options = { write: false };

const MANIFEST_PATH = './package.json';
const DEFAULT_TAG = 'latest';
const DEFAULT_TAG_PRERELEASE = 'next';
const NPM_BASE_URL = 'https://www.npmjs.com';

class npm extends Plugin {
  static isEnabled(options) {
    return hasAccess(MANIFEST_PATH) && options !== false;
  }

  constructor(...args) {
    super(...args);
    this.registerPrompts(prompts);
  }

  async init() {
    const { name, version: latestVersion, private: isPrivate, publishConfig } = require(path.resolve(MANIFEST_PATH));
    this.setContext({ name, latestVersion, private: isPrivate, publishConfig });

    const { publish, skipChecks } = this.options;

    const timeout = Number(this.options.timeout) * 1000;

    if (publish === false || isPrivate) return;

    if (skipChecks) return;

    const validations = Promise.all([this.isRegistryUp(), this.isAuthenticated(), this.getLatestRegistryVersion()]);

    await Promise.race([validations, rejectAfter(timeout, e(`Timed out after ${timeout}ms.`, docs))]);

    const [isRegistryUp, isAuthenticated, latestVersionInRegistry] = await validations;

    if (!isRegistryUp) {
      throw e(`Unable to reach npm registry (timed out after ${timeout}ms).`, docs);
    }

    if (!isAuthenticated) {
      throw e('Not authenticated with npm. Please `npm login` and try again.', docs);
    }

    if (!(await this.isCollaborator())) {
      const { username } = this.getContext();
      throw e(`User ${username} is not a collaborator for ${name}.`, docs);
    }

    if (!latestVersionInRegistry) {
      this.log.warn('No version found in npm registry. Assuming new package.');
    } else {
      if (!semver.eq(latestVersion, latestVersionInRegistry)) {
        this.log.warn(
          `Latest version in registry (${latestVersionInRegistry}) does not match package.json (${latestVersion}).`
        );
      }
    }
  }

  getName() {
    return this.getContext('name');
  }

  getLatestVersion() {
    return this.options.ignoreVersion ? null : this.getContext('latestVersion');
  }

  async bump(version) {
    const task = () =>
      this.exec(`npm version ${version} --no-git-tag-version`).catch(err => {
        if (/version not changed/i.test(err)) {
          this.log.warn(`Did not update version in package.json, etc. (already at ${version}).`);
        } else {
          return false;
        }
      });
    const tag = this.options.tag || (await this.resolveTag(version));
    this.setContext({ version, tag });
    return this.spinner.show({ task, label: 'npm version' });
  }

  release() {
    if (this.options.publish === false) return false;
    const publish = () => this.publish({ otpCallback });
    const otpCallback = this.config.isCI ? null : task => this.step({ prompt: 'otp', task });
    return this.step({ task: publish, label: 'npm publish', prompt: 'publish' });
  }

  isRegistryUp() {
    const registry = this.getRegistry();
    const registryArg = registry ? ` --registry ${registry}` : '';
    return this.exec(`npm ping${registryArg}`, { options }).then(
      () => true,
      err => {
        if (/code E40[04]|404.*(ping not found|No content for path)/.test(err)) {
          this.log.warn('Ignoring response from unsupported `npm ping` command.');
          return true;
        }
        return false;
      }
    );
  }

  isAuthenticated() {
    const registry = this.getRegistry();
    const registryArg = registry ? ` --registry ${registry}` : '';
    return this.exec(`npm whoami${registryArg}`, { options }).then(
      output => {
        const username = output ? output.trim() : null;
        this.setContext({ username });
        return true;
      },
      err => {
        this.debug(err);
        if (/code E40[04]/.test(err)) {
          this.log.warn('Ignoring response from unsupported `npm whoami` command.');
          return true;
        }
        return false;
      }
    );
  }

  isCollaborator() {
    const registry = this.getRegistry();
    const registryArg = registry ? ` --registry ${registry}` : '';
    const name = this.getName();
    const { username } = this.getContext();
    if (username === undefined) return true;
    if (username === null) return false;
    return this.exec(`npm access ls-collaborators ${name}${registryArg}`, { options }).then(
      output => {
        try {
          const collaborators = JSON.parse(output);
          const permissions = collaborators[username];
          return permissions && permissions.includes('write');
        } catch (err) {
          this.debug(err);
          return false;
        }
      },
      err => {
        this.debug(err);
        if (/code E400/.test(err)) {
          this.log.warn('Ignoring response from unsupported `npm access` command.');
          return true;
        }
        return /code E404/.test(err);
      }
    );
  }

  async getLatestRegistryVersion() {
    const registry = this.getRegistry();
    const registryArg = registry ? ` --registry ${registry}` : '';
    const name = this.getName();
    const latestVersion = this.getLatestVersion();
    const tag = await this.resolveTag(latestVersion);
    return this.exec(`npm show ${name}@${tag} version${registryArg}`, { options }).catch(() => null);
  }

  getRegistryPreReleaseTags() {
    return this.exec(`npm view ${this.getName()} dist-tags --json`, { options }).then(
      output => {
        try {
          const tags = JSON.parse(output);
          return Object.keys(tags).filter(tag => tag !== DEFAULT_TAG);
        } catch (err) {
          this.debug(err);
          return [];
        }
      },
      () => []
    );
  }

  getPackageUrl() {
    const baseUrl = this.getRegistry() || NPM_BASE_URL;
    return urlJoin(baseUrl, 'package', this.getName());
  }

  getRegistry() {
    const { publishConfig } = this.getContext();
    const registries = publishConfig
      ? publishConfig.registry
        ? [publishConfig.registry]
        : Object.keys(publishConfig)
            .filter(key => key.endsWith('registry'))
            .map(key => publishConfig[key])
      : [];
    return registries[0];
  }

  async guessPreReleaseTag() {
    const [tag] = await this.getRegistryPreReleaseTags();
    if (tag) {
      return tag;
    } else {
      this.log.warn(`Unable to get pre-release tag(s) from npm registry. Using "${DEFAULT_TAG_PRERELEASE}".`);
      return DEFAULT_TAG_PRERELEASE;
    }
  }

  async resolveTag(version) {
    const { tag } = this.options;
    const { isPreRelease, preReleaseId } = parseVersion(version);
    if (!isPreRelease) {
      return DEFAULT_TAG;
    } else {
      return tag || preReleaseId || (await this.guessPreReleaseTag());
    }
  }

  async publish({ otp = this.options.otp, otpCallback } = {}) {
    const { publishPath = '.' } = this.options;
    const { private: isPrivate, tag = DEFAULT_TAG } = this.getContext();
    const otpArg = otp ? `--otp ${otp}` : '';
    const dryRunArg = this.config.isDryRun ? '--dry-run' : '';
    if (isPrivate) {
      this.log.warn('Skip publish: package is private.');
      return false;
    }
    return this.exec(`npm publish ${publishPath} --tag ${tag} ${otpArg} ${dryRunArg}`, { options })
      .then(() => {
        this.setContext({ isReleased: true });
      })
      .catch(err => {
        this.debug(err);
        if (/one-time pass/.test(err)) {
          if (otp != null) {
            this.log.warn('The provided OTP is incorrect or has expired.');
          }
          if (otpCallback) {
            return otpCallback(otp => this.publish({ otp, otpCallback }));
          }
        }
        throw err;
      });
  }

  afterRelease() {
    const { isReleased } = this.getContext();
    if (isReleased) {
      this.log.log(`🔗 ${this.getPackageUrl()}`);
    }
  }
}

module.exports = npm;

==> release-it-14.11.0-next.0.tgz: package/lib/plugin/npm/npm.js <==
const path = require('path');
const semver = require('semver');
const urlJoin = require('url-join');
const Plugin = require('../Plugin');
const { hasAccess, rejectAfter, parseVersion, e } = require('../../util');
const prompts = require('./prompts');

const docs = 'https://git.io/release-it-npm';

const options = { write: false };

const MANIFEST_PATH = './package.json';
const LOCKFILE_YARN = './yarn.lock';
const LOCKFILE_PNPM = './pnpm-lock.yaml';
const DEFAULT_TAG = 'latest';
const DEFAULT_TAG_PRERELEASE = 'next';
const NPM_BASE_URL = 'https://www.npmjs.com';

const fixArgs = args => (args ? (typeof args === 'string' ? args.split(' ') : args) : []);

const programs = {
  npm: {
    version: ({ version }) => `npm version ${version} --no-git-tag-version`,
    publish: ({ publishPath, tag, otpArg, dryRunArg, publishArgs }) => {
      const args = [publishPath, `--tag ${tag}`, otpArg, dryRunArg, ...fixArgs(publishArgs)].filter(Boolean);
      return `npm publish ${args.join(' ')}`;
    }
  },
  pnpm: {
    version: ({ version }) => `npm version ${version} --no-git-tag-version`,
    publish: ({ publishPath, tag, publishArgs }) => {
      const args = [...fixArgs(publishArgs), publishPath, `--tag ${tag}`].filter(Boolean);
      return `pnpm publish ${args.join(' ')}`;
    }
  },
  yarn: {
    version: ({ version }) => `yarn version ${version} --no-git-tag-version`,
    publish: ({ tag, publishArgs }) => {
      const args = [`--tag ${tag}`, ...fixArgs(publishArgs)].filter(Boolean);
      return `yarn npm publish ${args.join(' ')}`;
    }
  }
};

class npm extends Plugin {
  static isEnabled(options) {
    return hasAccess(MANIFEST_PATH) && options !== false;
  }

  constructor(...args) {
    super(...args);
    this.registerPrompts(prompts);
  }

  getInitialOptions(options, namespace) {
    options[namespace] = options[namespace] || {};
    let program;
    switch (true) {
      case options[namespace].yarn === true:
        program = 'yarn';
        break;
      case options[namespace].pnpm === true:
        program = 'pnpm';
        break;
      case hasAccess(LOCKFILE_YARN) && options[namespace].yarn !== false:
        program = 'yarn';
        break;
      case hasAccess(LOCKFILE_PNPM) && options[namespace].pnpm !== false:
        program = 'pnpm';
        break;
      default:
        program = 'npm';
        break;
    }
    options[namespace].program = program;
    return options[namespace];
  }

  async init() {
    const { name, version: latestVersion, private: isPrivate, publishConfig } = require(path.resolve(MANIFEST_PATH));
    this.setContext({ name, latestVersion, private: isPrivate, publishConfig });

    const { publish, skipChecks } = this.options;

    const timeout = Number(this.options.timeout) * 1000;

    if (publish === false || isPrivate) return;

    if (skipChecks) return;

    const validations = Promise.all([this.isRegistryUp(), this.isAuthenticated(), this.getLatestRegistryVersion()]);

    await Promise.race([validations, rejectAfter(timeout, e(`Timed out after ${timeout}ms.`, docs))]);

    const [isRegistryUp, isAuthenticated, latestVersionInRegistry] = await validations;

    if (!isRegistryUp) {
      throw e(`Unable to reach npm registry (timed out after ${timeout}ms).`, docs);
    }

    if (!isAuthenticated) {
      throw e('Not authenticated with npm. Please `npm login` and try again.', docs);
    }

    if (!(await this.isCollaborator())) {
      const { username } = this.getContext();
      throw e(`User ${username} is not a collaborator for ${name}.`, docs);
    }

    if (!latestVersionInRegistry) {
      this.log.warn('No version found in npm registry. Assuming new package.');
    } else {
      if (!semver.eq(latestVersion, latestVersionInRegistry)) {
        this.log.warn(
          `Latest version in registry (${latestVersionInRegistry}) does not match package.json (${latestVersion}).`
        );
      }
    }
  }

  getCommand(command, options) {
    if (command in programs[this.options.program]) {
      return programs[this.options.program][command](options);
    } else {
      return `npm ${command}`;
    }
  }

  getName() {
    return this.getContext('name');
  }

  getLatestVersion() {
    return this.options.ignoreVersion ? null : this.getContext('latestVersion');
  }

  async bump(version) {
    const tag = this.options.tag || (await this.resolveTag(version));
    this.setContext({ version, tag });

    if (this.config.isUpdate) return false;

    const command = this.getCommand('version', { version });
    const task = () => this.exec(command);
    const label = command.split(' ').slice(0, 2).join(' ');
    return this.spinner.show({ task, label });
  }

  release() {
    if (this.options.publish === false) return false;
    const publish = () => this.publish({ otpCallback });
    const otpCallback = this.config.isCI ? null : task => this.step({ prompt: 'otp', task });
    const command = `${this.options.program} publish`;
    return this.step({ task: publish, label: command, prompt: 'publish' });
  }

  isRegistryUp() {
    const registry = this.getRegistry();
    const registryArg = registry ? ` --registry ${registry}` : '';
    return this.exec(`npm ping${registryArg}`, { options }).then(
      () => true,
      err => {
        if (/code E40[04]|404.*(ping not found|No content for path)/.test(err)) {
          this.log.warn('Ignoring response from unsupported `npm ping` command.');
          return true;
        }
        return false;
      }
    );
  }

  isAuthenticated() {
    const registry = this.getRegistry();
    const registryArg = registry ? ` --registry ${registry}` : '';
    return this.exec(`npm whoami${registryArg}`, { options }).then(
      output => {
        const username = output ? output.trim() : null;
        this.setContext({ username });
        return true;
      },
      err => {
        this.debug(err);
        if (/code E40[04]/.test(err)) {
          this.log.warn('Ignoring response from unsupported `npm whoami` command.');
          return true;
        }
        return false;
      }
    );
  }

  isCollaborator() {
    const registry = this.getRegistry();
    const registryArg = registry ? ` --registry ${registry}` : '';
    const name = this.getName();
    const { username } = this.getContext();
    if (username === undefined) return true;
    if (username === null) return false;
    return this.exec(`npm access ls-collaborators ${name}${registryArg}`, { options }).then(
      output => {
        try {
          const collaborators = JSON.parse(output);
          const permissions = collaborators[username];
          return permissions && permissions.includes('write');
        } catch (err) {
          this.debug(err);
          return false;
        }
      },
      err => {
        this.debug(err);
        if (/code E400/.test(err)) {
          this.log.warn('Ignoring response from unsupported `npm access` command.');
        } else {
          this.log.warn(`Unable to verify if user ${username} is a collaborator for ${name}.`);
        }
        return true;
      }
    );
  }

  async getLatestRegistryVersion() {
    const registry = this.getRegistry();
    const registryArg = registry ? ` --registry ${registry}` : '';
    const name = this.getName();
    const latestVersion = this.getLatestVersion();
    const tag = await this.resolveTag(latestVersion);
    return this.exec(`npm show ${name}@${tag} version${registryArg}`, { options }).catch(() => null);
  }

  getRegistryPreReleaseTags() {
    return this.exec(`npm view ${this.getName()} dist-tags --json`, { options }).then(
      output => {
        try {
          const tags = JSON.parse(output);
          return Object.keys(tags).filter(tag => tag !== DEFAULT_TAG);
        } catch (err) {
          this.debug(err);
          return [];
        }
      },
      () => []
    );
  }

  getPackageUrl() {
    const baseUrl = this.getRegistry() || NPM_BASE_URL;
    return urlJoin(baseUrl, 'package', this.getName());
  }

  getRegistry() {
    const { publishConfig } = this.getContext();
    const registries = publishConfig
      ? publishConfig.registry
        ? [publishConfig.registry]
        : Object.keys(publishConfig)
            .filter(key => key.endsWith('registry'))
            .map(key => publishConfig[key])
      : [];
    return registries[0];
  }

  async guessPreReleaseTag() {
    const [tag] = await this.getRegistryPreReleaseTags();
    if (tag) {
      return tag;
    } else {
      this.log.warn(`Unable to get pre-release tag(s) from npm registry. Using "${DEFAULT_TAG_PRERELEASE}".`);
      return DEFAULT_TAG_PRERELEASE;
    }
  }

  async resolveTag(version) {
    const { tag } = this.options;
    const { isPreRelease, preReleaseId } = parseVersion(version);
    if (!isPreRelease) {
      return DEFAULT_TAG;
    } else {
      return tag || preReleaseId || (await this.guessPreReleaseTag());
    }
  }

  async publish({ otp = this.options.otp, otpCallback } = {}) {
    const { program, publishPath = '.', publishArgs, isDryRun } = this.options;
    const { private: isPrivate, tag = DEFAULT_TAG } = this.getContext();
    const otpArg = otp ? `--otp ${otp}` : '';
    const dryRunArg = isDryRun ? '--dry-run' : '';
    if (isPrivate) {
      this.log.warn('Skip publish: package is private.');
      return false;
    }
    const command = this.getCommand('publish', { publishPath, tag, otpArg, dryRunArg, publishArgs });
    return this.exec(command, { options: { write: program !== 'npm' } })
      .then(() => {
        this.setContext({ isReleased: true });
      })
      .catch(err => {
        this.debug(err);
        if (/one-time pass/.test(err)) {
          if (otp != null) {
            this.log.warn('The provided OTP is incorrect or has expired.');
          }
          if (otpCallback) {
            return otpCallback(otp => this.publish({ otp, otpCallback }));
          }
        }
        throw err;
      });
  }

  afterRelease() {
    const { isReleased } = this.getContext();
    if (isReleased) {
      this.log.log(`🔗 ${this.getPackageUrl()}`);
    }
  }
}

module.exports = npm;
